/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** proto.h - Function declarations
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: proto.h,v 1.173 2002/12/17 14:29:57 wcampbel Exp $
*/

#ifndef PROTO_H
#define PROTO_H

#include "stats.h"

/* Macros that were in m_stats.h */
#define DecreaseOpers(x)        { (x)->opers--; netinfo.opers--; }
#define DecreaseUsers(x)        { (x)->users--; netinfo.users--; }
#define DecreaseServers()       netinfo.servers--;
#define DecreaseChannels()	netinfo.channels--;

/* Macros to replace ugly inlines */
#define msgtoken(x,y)		generictoken(' ', MAXPARA, (x), (y))
#define tabtoken(x,y)		generictoken('\t', MAXPARA, (x), (y))

/* Fixes for broken operating systems */
#ifndef HAVE_STRLCAT
size_t strlcat(char *, const char *, size_t);
#endif

#ifndef HAVE_STRLCPY
size_t strlcpy(char *, const char *, size_t);
#endif

/* Simplification macros */
#define IsPrivmsg()	(settings.flags & FLAG_PRIVMSG)
#define IsOnlyOpers()	(settings.flags & FLAG_ONLYOPERS)
#define IsAOnlyOpers()  (settings.flags & FLAG_AONLYOPERS)
#define IsHtml()	(settings.flags & FLAG_HTML)
#define IsDump()	(settings.flags & FLAG_DUMP)
#define IsSetOper()	(settings.flags & FLAG_SETOPER)
#define IsFlood()	(settings.flags & FLAG_FLOOD)
#define IsVHost()	(settings.flags & FLAG_VHOST)
#define IsSilent()	(settings.flags & FLAG_SILENT)
#define IsNotifyEOB()	(settings.flags & FLAG_NOTIFYEOB)
#define IsUnauthChat()	(settings.flags & FLAG_UNAUTHCHAT)
#define IsDCCOperOnly()	(settings.flags & FLAG_DCCOPERONLY)
#define IsPubDelayed()	(settings.flags & FLAG_PUBDELAYED)
#define IsOper(u)	((u)->flags & UFLAG_ISOPER)
#define IsAway(u)	((u)->flags & UFLAG_ISAWAY)
#define IsFloodSent(u)	((u)->flags & UFLAG_FLOODSENT)
#define IsOperOn(myu)	((myu)->flags & MFLAG_ISON)
#define IsOperDCC(myu)	((myu)->flags & MFLAG_DCC)
#define IsDebug()	(me.cmdflags & CMD_DEBUG)
#define IsNoFork()	(me.cmdflags & CMD_NOFORK)
#define IsSHide(s)	((s)->flags & SFLAG_HIDE)

#define IsSPPrivmsg()	(sp.flags & SPF_PRIVMSG)
#define IsSPChanmsg()	(sp.flags & SPF_CHANMSG)
#define IsSPEnableText()	(sp.flags & SPF_ENABLE_TEXT)
#define IsSPEnableExpire()	(sp.flags & SPF_EXPIRE)
#define IsSPReport()	(sp.flags & SPF_REPORT)
#define IsSPHide()	(sp.flags & SPF_HIDE)

#define SetEOB(s)	((s)->flags |= SFLAG_EOB)
#define IsSetEOB(s)	((s)->flags & SFLAG_EOB)

/* External variables... */
extern int servsock;
extern struct Connection *sconn;
extern int need_rehash;
#ifdef SPLITSERV
extern FILE* split_log;
#endif
extern FILE* main_log;
#ifdef IO_LOG
extern FILE* io_log;
#endif

extern const char version[];

#ifdef SUPPORT_MSG_UH
#ifdef STATSERV
extern char s_UH[MAXUSER + MAXHOST + 2];
#endif
#ifdef SPLITSERV
extern char sp_UH[MAXUSER + MAXHOST + 2];
#endif
#ifdef JUPESERV
extern char j_UH[MAXUSER + MAXHOST + 2];
#endif
extern char a_UH[MAXUSER + MAXHOST + 2];
#endif

#ifdef SUPPORT_MSG_OPERS
extern char s_OPERS[MAXHOST + 7];
#endif

#if defined(SPLITSERV)
int is_delayed;
#endif
#if defined(SPYMODE) || defined(ALERTMODE)
int ad_delayed;
#endif

#ifdef SPLITSERV
struct SpServer* map_head;
struct SpServer* map_uplink;
dlink_list split_table;
dlink_list server_table;

/* split-util.c */
void dump_map();
void html_map(FILE *, struct SpServer*, int, int *);
struct mapreturn PrintMap(char*, struct SpServer*, int, int *);
void PrintSplit(char*);
void PrintDump(char *);
void expire_splits(void);

/* split-data.c */
void SpInitTables();
struct SpServer* SpNewServer(char *, char *);
struct SpServer* SpFindServer(char *);
void SpDelServer(char *);
void SpNewSplit(char *);
void SpDelSplit(char *);
struct SpServer* SpFindSplit(char *);

/* splitserv.c */
void splitserv(struct User *, char *, char *);

#endif

/* main.c */
void ServTermination(void);

/* adminserv.c */
void adminserv(struct User *, char *);

/* sock.c */
int ConnectTo(char *, int);
void read_loop();
void close_dccs(void);

/* send.c */
int sts(char *, ...);
int sendto_dcc(int, char *, ...);
void sendto_all_dcc(char *, ...);
void sendto_partyline(struct Connection *, char *, ...);
void sendto_dcc_flag(int, char *, ...);
void send_dcc_help_file(int, char *, char *);

/* util.c */
char *timediff(time_t);
void log(char *, ...);
#ifdef SPLITSERV
void splog(char *, ...);
#endif
#ifdef IO_LOG
void iolog(char *, ...);
#endif
void ResetLogs();
char *sctime(time_t);
char *sftime(time_t);
char *irc_time(time_t);
#if defined(CRYPT) || defined(MD5)
char *make_salt();
#endif /* CRYPT || MD5 */
int tokenize(char *, char **);
int generictoken(char, int, char *, char **);
#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_HYBSVS) || defined(IRCD_BAHAMUT)
int sjtoken(char *, char **);
#endif
void *smalloc(size_t);
unsigned long HASH(const unsigned char *, int);
char *VersionFlags();
#ifdef OPER_MODES
char opermode(struct User *);
#endif
void SigTerm(int);
void SigUsr1(int);
void SigHup(int);
void SigInt(int);
int is_channel(char *);
void alert(char *, ...);
void spy(char *, ...);
void debug(char *, ...);
void InitLists(void);
void ExpandParv(char *, int, int, int, char **);
int is_midnight(void);
void select_delay(void);

/* conf.c */
void strip(char *);
void InitDefaults();
void VerifyConf();
void ConfLoad();
void rehash();

/* init.c */
int login();
#ifdef STATSERV
void init_statserv();
#endif
#ifdef STATSERV
void init_join();
#endif
void init_admin();
#ifdef SPLITSERV
void init_splitserv();
void init_spjoin();
#endif
#ifdef JUPESERV
void init_jupeserv();
#endif
#ifdef SPYMODE
void init_spy_join();
#endif
#ifdef ALERTMODE
void init_alert_join();
#endif
void InitMessageStrings();
#ifdef PUBSTATS
void init_pubjoin();
#endif

/* ircd.c */
void InitMsgHash(void);
void parse();
void privmsg(char *, const char *, int, char *, ...);
void send_help_file(char *, char *, int, char *, char *);
void wallops(char *, char *, ...);
int flood(struct User *);
void SendQuit(struct User *, int, char *);

/* statserv.c */
#ifdef STATSERV
#ifdef IRCD_2_10
void statserv(struct User *, char *, char *, int);
#else
void statserv(struct User *, char *, char *);
#endif
#endif

/* servers.c */
struct Server *serverlist[S_TABLE_SIZE];
#ifdef IRCD_2_10
void AddServer(char *, int, int, char *, char *);
struct Server *findservtoken(int);
#else
void AddServer(char *, int, char *, char *);
#endif
void FlushServer(char *);
void DelServer(char *);
struct Server *findserver(char *);
void init_server_hash();
void IncreaseServers();

/* channels.c */
struct Channel *chanlist[C_TABLE_SIZE];
#ifdef IRCD_2_10
struct Channel *AddChannel(char *);
#else
struct Channel *AddChannel(char *, long);
#endif
void DelChannel(struct Channel *);
struct Channel *findchannel(char *);
void init_channel_hash();
void IncreaseChannels();
void ChannelMode(struct Channel *, int, char **);
char *ModeString(struct Channel *);
void AddUserToChannel(struct Channel *, char *);
void AddUserFlagToChannel(struct Channel *, struct User *, unsigned int);
char *CUModeString(struct ChanUser *);
struct ChanUser *findchanuser(struct Channel *, struct User *);
void DelUserFromChannel(struct Channel *, struct User *);
void UnlinkChannelFromUser(struct Channel *, struct User *);

/* services.c */
#ifdef IRCD_2_10
dlink_list servicehead;
struct Service *findservice(char *, char *);
void AddService(char *, char *, char *, char *);
void DelService(char *, char *);
void DelServicesByServer(char *);
#endif

/* users.c */
struct User *userlist[U_TABLE_SIZE];
#ifdef IRCD_2_10
void AddUser(char *, char *, char *, struct Server *, char *);
#else
void AddUser(char *, char *, char *, char *, char *);
#endif
void DelUser(char *);
#if defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
void Change_User(struct User *, char *, int);
#else
void Change_User(struct User *, char *);
#endif
#ifdef SUPPORT_HOSTSERV
void Change_Host(struct User *, char *);
void Change_Ident(struct User *, char *);
#endif
struct User *finduser(char *);
void UserMode(char *, char *);
void init_user_hash();
void IncreaseUsers(struct Server *);

/* myusers.c */
dlink_list myuhead;
void LoadMyUsers();
void SaveMyUsers();
void DeleteMyUser(char *);
struct MyUser *findmyuser(char *);

/* tld.c */
struct TLD *tldlist[T_TABLE_SIZE];
struct TLD *findtld(char *);
struct TLD *AddTLD(struct User *);
void LoadTLD();
void init_tld();

/* stats.c */
dlink_list shead;
dlink_list hiddenhead;
void IncreaseAway(struct Server *);
void IncreaseOpers(struct Server *);
void AddStats(struct Server *);
struct Stats *findstats(char *);
void SaveStats();
void LoadStats();
void AddHidden(char *, int);
#if 0
void DelHidden(char *);
#endif
struct Hidden* findhidden(char *);

/* html.c */
void DumpHTML();

/* datadump.c */
void datadump();

#ifdef JUPESERV

dlink_list jupehead;

/* jupe.c */
void LoadJupes();
void SaveJupes();
void AddJupe(char *, char *, long, char *);
void InitJupe(struct Jupe *);
void DelJupe(char *);
struct Jupe* FindJupe(char *);
void FlushJupeList();

/* jupeserv.c */
void jupeserv(struct User *, char *);

#endif

/* info.c - autogenerated */
char *stats_version;
char *stats_date;

/* evh.c */
void StartInitialEvents(void);
void FlushEvents(void);
void ResetTimedEvents(void);
#if defined(SPYMODE) || defined(ALERTMODE)
/* To deal with KILL's against the administration service, this needs to
** be exported
*/
void evh_CheckAdDelayed(void *unused);
#endif
void evh_ConnectServer(void *unused);

/* operlist.c */
dlink_list global_oper_list;
void AddOperToList(dlink_list *, struct User *);
void DelOperFromList(dlink_list *, struct User *);

/* connection.c */
dlink_list connection_list;
struct Connection *add_connection(int, char *, unsigned int);
void delete_connection(int);
struct Connection *find_connection(int);

/* dcc.c */
void dcc_parse(int, char *);

/* capab.c */
#ifndef IRCD_2_10
char *generate_capab_string(void);
#endif

/* htmlsup.c */
void InitHTMLArray(void);
char *HTMLReplace(char *);

/* pubstats.c */
#ifdef PUBSTATS
void pubstats(struct User *, char *, char *);
#endif

#endif
