/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** connection.c - Socket struct manipulation routines
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: connection.c,v 1.7 2002/11/08 15:21:08 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "dlink.h"
#include "proto.h"

dlink_list connection_list;

struct Connection *add_connection(int sock, char *name, unsigned int type)
{
	struct Connection *c = smalloc(sizeof(struct Connection));
	dlink_node *dl = dlink_create();

	c->fd = sock;
	c->flags |= type;
	strlcpy(c->name, name, LINE);
	dlink_add(c, dl, &connection_list);
	memcounts.connection++;

	return c;
}

void delete_connection(int sock)
{
	struct Connection *c;
	dlink_node *dl;

	DLINK_FOREACH(dl, connection_list.head)
	{
		c = (struct Connection *) dl->data;
		if (c->fd == sock)
		{
			/* Mark this connection as no longer online in
			** case it was logged in
			*/
			if (c->myuser != NULL)
				c->myuser->flags &= ~MFLAG_DCC;
			dlink_delete(dl, &connection_list);
			free(c);
			memcounts.connection--;
			dlink_free(dl);
			return;
		}
	}

	log("Unable to delete connection struct:  fd %d", sock);
}

struct Connection *find_connection(int sock)
{
	struct Connection *c;
	dlink_node *dl;

	DLINK_FOREACH(dl, connection_list.head)
	{
		c = (struct Connection *) dl->data;
		if (c->fd == sock)
			return c;
	}
	return NULL;
}
