/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** htmlsup.c - Support functions to the HTML dump.
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: htmlsup.c,v 1.3 2002/11/12 18:43:31 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"

#define SINGLE		0
#define START_NEEDED	1
#define END_NEEDED	2

/* Certain characters have to be escaped when put into an HTML file.
** These characters often are replaced by longer strings, so a simple
** character replacement will not work.
*/

struct StringReplace
{
	char character;
	int type;
	int len;
	char *replacement;
	int len2;
	char *repl2;
};

static struct StringReplace replace_list[] =
{
	{ '\002',	START_NEEDED,	3,"<B>",	4,"</B>" },
	{ '\037',	START_NEEDED,	3,"<U>",	4,"</U>" },
	{ '<',		SINGLE,		4,"&lt;",	0,NULL },
	{ '>',		SINGLE,		4,"&gt;",	0,NULL },
	{ '\"',		SINGLE,		6,"&quot;",	0,NULL },
	{ '&',		SINGLE,		5,"&amp;",	0,NULL },
	{ ' ',		SINGLE,		0,NULL,		0,NULL }
};

static struct StringReplace *replace_array[256];

void InitHTMLArray(void)
{
	int i;

	for (i = 0; i < 256; i++)
		replace_array[i] = NULL;

	for (i = 0; replace_list[i].replacement != NULL; i++)
		replace_array[(int) replace_list[i].character] =
			&replace_list[i];
}

char *HTMLReplace(char *source)
{
	int i;
	int j = 0;
	static char dest[BUFSIZE];

	if (source == NULL)
		return NULL;

	dest[0] = '\0';

	for (i = 0; source[i] != '\0'; i++)
	{
		if (j == BUFSIZE)
		{
			dest[BUFSIZE - 1] = '\0';
			return dest;
		}
		if (replace_array[(int) source[i]] != NULL)
		{
		    if (replace_array[(int) source[i]]->type == SINGLE)
		    {
			dest[j] = '\0';
			strlcat(dest,
				replace_array[(int) source[i]]->replacement,
				BUFSIZE);
			j += replace_array[(int) source[i]]->len;
		    }
		    else if (replace_array[(int) source[i]]->type ==
				START_NEEDED)
		    {
			dest[j] = '\0';
			strlcat(dest,
				replace_array[(int) source[i]]->replacement,
				BUFSIZE);
			j += replace_array[(int) source[i]]->len;
			replace_array[(int) source[i]]->type = END_NEEDED;
		    }
		    else if (replace_array[(int) source[i]]->type ==
				END_NEEDED)
		    {
			dest[j] = '\0';
			strlcat(dest,
				replace_array[(int) source[i]]->repl2,
				BUFSIZE);
			j += replace_array[(int) source[i]]->len2;
			replace_array[(int) source[i]]->type = START_NEEDED;
		    }
		}
		else
		{
			dest[j] = source[i];
			j++;
		}
	}

	if (j == BUFSIZE)
		dest[BUFSIZE - 1] = '\0';
	else
		dest[j] = '\0';

	return dest;
}
