/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** init.c - Stats initialization functions
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: init.c,v 1.28 2002/12/17 14:29:59 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"

#ifdef SUPPORT_MSG_UH
#ifdef STATSERV
char s_UH[MAXUSER + MAXHOST + 2];
#endif
#ifdef SPLITSERV
char sp_UH[MAXUSER + MAXHOST + 2];
#endif
#ifdef JUPESERV
char j_UH[MAXUSER + MAXHOST + 2];
#endif
char a_UH[MAXUSER + MAXHOST + 2];
#endif

#ifdef SUPPORT_MSG_OPERS
/* 7 is "opers@" and the \0 terminator
** I should PROBABLY force this to only be enabled on Hybrid 6
*/
char s_OPERS[MAXHOST + 7];
#endif

/* These are all horrible functions, peppered with ifdef's */

/* This is ugly, logging into each server type is slightly different */
int login()
{
	int ret; /* Check for write errors */

#if defined(IRCD_HYBRID)
	ret = sts("PASS %s :TS", settings.pass);
	if (ret == 1)
		return 1;
	sts("SERVER %s 1 :%s", settings.name, settings.desc);
	sts("SVINFO 3 1 0 :%ld", time(NULL));
#elif defined(IRCD_HYB6) || defined(IRCD_HYB7) || defined(IRCD_HYBSVS)
	ret = sts("PASS %s :TS", settings.pass);
	if (ret == 1)
		return 1;
	sts("CAPAB :%s", me.capabs);
        sts("SERVER %s 1 :%s", settings.name, settings.desc);
        sts("SVINFO 5 3 0 :%ld", time(NULL));
#elif defined(IRCD_2_10)
	ret = sts("PASS %s :021030003 IRC|%sHst P", settings.pass,
		IsDebug() ? "D" : "");
	if (ret == 1)
		return 1;
	sts("SERVER %s 1 :%s", settings.name, settings.desc);
#elif defined(IRCD_BAHAMUT)
	ret = sts("PASS %s :TS", settings.pass);
	if (ret == 1)
		return 1;
	/* Using SSJOIN will save bandwidth, UNCONNECT is required */
	sts("CAPAB :%s", me.capabs);
	sts("SERVER %s 1 :%s", settings.name, settings.desc);
	sts("SVINFO 3 1 0 :%ld", time(NULL));
#endif
	return 0;
}

#ifdef JUPESERV

/* Bring up the jupe service */
void init_jupeserv()
{
        struct User *u = NULL;

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_HYBSVS)
        sts("NICK %s 1 1 +io %s %s %s :Jupe Service",
                JupeServ.nick, JupeServ.user, JupeServ.host, settings.name);
#elif defined(IRCD_2_10)
        sts("NICK %s 1 %s %s %d +io :Jupe Service", JupeServ.nick,
               JupeServ.user, JupeServ.host, 1);
#elif defined(IRCD_BAHAMUT)
	sts("NICK %s 1 1 +io %s %s %s 1 :Jupe Service",
		JupeServ.nick, JupeServ.user, JupeServ.host, settings.name);
#endif

#ifdef IRCD_2_10
        AddUser(JupeServ.nick, JupeServ.user, JupeServ.host,
                findserver(settings.name), "Jupe Service");
#else
        AddUser(JupeServ.nick, JupeServ.user, JupeServ.host, settings.name,
		"Jupe Service");
#endif

        u = finduser(JupeServ.nick);
        IncreaseOpers(findserver(settings.name));
        u->flags |= UFLAG_ISOPER;
#ifdef OPER_MODES
        u->operflags |= OPER_MYOPER;
#endif
	AddOperToList(&global_oper_list, u);
	JupeServ.u = u;
}

#endif

#ifdef STATSERV
/* Init the StatServ client */
void init_statserv()
{
	struct User *u;

#ifdef SUPPORT_MSG_UH
	snprintf(StatServ.rname, LINE, "/msg %s HELP", s_UH);
#else
	snprintf(StatServ.rname, LINE, "/msg %s HELP", StatServ.nick);
#endif

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_HYBSVS)
	sts("NICK %s 1 1 %s %s %s %s :%s", StatServ.nick,
		IsSetOper() ? "+o" : "+", StatServ.user, StatServ.host,
		settings.name, StatServ.rname);
#elif defined(IRCD_2_10)
	sts("NICK %s 1 %s %s %d %s :%s", StatServ.nick,
		StatServ.user, StatServ.host, 
		/* I am Server ID 1 */
		1, IsSetOper() ? "+o" : "+", StatServ.rname);
#elif defined(IRCD_BAHAMUT)
	sts("NICK %s 1 1 %s %s %s %s 1 :%s", StatServ.nick,
		IsSetOper() ? "+o" : "+", StatServ.user, StatServ.host,
		settings.name, StatServ.rname);
#endif

#ifdef IRCD_2_10
	AddUser(StatServ.nick, StatServ.user, StatServ.host,
		findserver(settings.name), StatServ.rname);
#else
	AddUser(StatServ.nick, StatServ.user, StatServ.host, settings.name,
		StatServ.rname);
#endif

/* If the StatServ client is an oper, increase oper count here
** and set it as an oper
*/
	u = finduser(StatServ.nick);
	if (IsSetOper()) {
		IncreaseOpers(findserver(settings.name));
		u->flags |= UFLAG_ISOPER;
#ifdef OPER_MODES
		u->operflags |= OPER_MYOPER;
#endif
		AddOperToList(&global_oper_list, u);
	}
	StatServ.u = u;
}
#endif /* STATSERV */

#ifdef SPLITSERV
/* Bring up the Split detecting/reporting client if it was enabled */
void init_splitserv()
{
	struct User *u;

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_HYBSVS)
        sts("NICK %s 1 1 + %s %s %s :%s", SplitServ.nick, SplitServ.user,
                SplitServ.host, settings.name, sp.desc);
#elif defined(IRCD_2_10)
        sts("NICK %s 1 %s %s %d + :%s", SplitServ.nick, SplitServ.user,
                SplitServ.host, 1, sp.desc);
                /* This is the server token, myself is 1 */
#elif defined(IRCD_BAHAMUT)
	sts("NICK %s 1 1 + %s %s %s 1 :%s", SplitServ.nick, SplitServ.user,
		SplitServ.host, settings.name, sp.desc);
#endif

#ifdef IRCD_2_10
        AddUser(SplitServ.nick, SplitServ.user, SplitServ.host,
		findserver(settings.name), sp.desc);
#else
        AddUser(SplitServ.nick, SplitServ.user, SplitServ.host, settings.name,
		sp.desc);
#endif
	u = finduser(SplitServ.nick);
	SplitServ.u = u;
}
#endif

/* Bring up the administration service */
void init_admin()
{
	struct User *u = NULL;

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_HYBSVS)
        sts("NICK %s 1 1 +io %s %s %s :Administration Service",
		AdminServ.nick, AdminServ.user, AdminServ.host, settings.name);
#elif defined(IRCD_2_10)
        sts("NICK %s 1 %s %s %d +io :Administration Service", AdminServ.nick,
		AdminServ.user, AdminServ.host, 1);
#elif defined(IRCD_BAHAMUT)
	sts("NICK %s 1 1 +io %s %s %s 1 :Administration Service",
		AdminServ.nick, AdminServ.user, AdminServ.host, settings.name);
#endif

#ifdef IRCD_2_10
        AddUser(AdminServ.nick, AdminServ.user, AdminServ.host,
		findserver(settings.name), "Administration Service");
#else
        AddUser(AdminServ.nick, AdminServ.user, AdminServ.host, settings.name,
		"Administration Service");
#endif

	u = finduser(AdminServ.nick);
	IncreaseOpers(findserver(settings.name));
	u->flags |= UFLAG_ISOPER;
#ifdef OPER_MODES
        u->operflags |= OPER_MYOPER;
#endif
	AddOperToList(&global_oper_list, u);
	AdminServ.u = u;
}

#ifdef STATSERV
/* Join the channel and obtain ops if it is a new channel
** We don't take over channels with Stats
*/
void init_join()
{
   dlink_node *dl;
   struct Channel *c;
#ifndef IRCD_2_10
   time_t now = time(NULL);
#endif

   DLINK_FOREACH(dl, me.chan_list.head)
   {
#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
	sts(":%s SJOIN %ld %s + :@%s", settings.name, 
	    now, (char *) dl->data, StatServ.nick);
        c = findchannel((char *) dl->data);
        if (c == NULL)
                c = AddChannel((char *) dl->data, now);
#elif defined(IRCD_2_10)
	/* Here is the only one that we hack ops... */
	sts(":%s NJOIN %s :@%s", settings.name, (char *) dl->data, 
		StatServ.nick);
	c = findchannel((char *) dl->data);
	if (c == NULL)
		c = AddChannel((char *) dl->data);
#endif
	AddUserFlagToChannel(c, StatServ.u, CMODE_OP);
   }
}
#endif

#ifdef SPLITSERV
/* Join the channel and obtain ops if it is a new channel
** We don't take over channels with Split
*/
void init_spjoin()
{
	struct Channel *c;

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
	time_t now = time(NULL);

	sts(":%s SJOIN %ld %s + :@%s", settings.name,
        	time(NULL), sp.join_chan, SplitServ.nick);
        c = findchannel(sp.join_chan);
        if (c == NULL)
                c = AddChannel(sp.join_chan, now);
#elif defined(IRCD_2_10)
        /* Here is the only one that we hack ops... */
        sts(":%s NJOIN %s :@%s", settings.name, sp.join_chan, SplitServ.nick);
        c = findchannel(sp.join_chan);
        if (c == NULL)
                c = AddChannel(sp.join_chan);
#endif
	AddUserFlagToChannel(c, SplitServ.u, CMODE_OP);
}
#endif

#ifdef PUBSTATS
/* Join the channel and obtain ops if it is a new channel
** We don't take over channels with Stats
*/
void init_pubjoin()
{
	struct Channel *c;

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
	time_t now = time(NULL);

        sts(":%s SJOIN %ld %s + :@%s", settings.name,
                now, settings.pub_chan, StatServ.nick);
        c = findchannel(settings.pub_chan);
        if (c == NULL)
                c = AddChannel(settings.pub_chan, now);
#elif defined(IRCD_2_10)
        /* Here is the only one that we hack ops... */
        sts(":%s NJOIN %s :@%s", settings.name, settings.pub_chan,
		StatServ.nick);
        c = findchannel(settings.pub_chan);
        if (c == NULL)
                c = AddChannel(settings.pub_chan);
#endif
	AddUserFlagToChannel(c, StatServ.u, CMODE_OP);
}
#endif

#ifdef ALERTMODE
/* Join the channel and obtain ops if it is a new channel
** We don't take over channels with AdminServ...
** Though in this case we might want to.....
*/
void init_alert_join()
{
	struct Channel *c;

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
	time_t now = time(NULL);

        sts(":%s SJOIN %ld %s + :@%s", settings.name,
                now, ALERTCHAN, AdminServ.nick);
        c = findchannel(ALERTCHAN);
        if (c == NULL)
                c = AddChannel(ALERTCHAN, now);
#elif defined(IRCD_2_10)
        /* Here is the only one that we hack ops... */
        sts(":%s NJOIN %s :@%s", settings.name, ALERTCHAN, AdminServ.nick);
        c = findchannel(ALERTCHAN);
        if (c == NULL)
                c = AddChannel(ALERTCHAN);
#endif
	AddUserFlagToChannel(c, AdminServ.u, CMODE_OP);
}
#endif

#ifdef SPYMODE
/* Join the channel and obtain ops if it is a new channel
** We don't take over channels with AdminServ...
** Though in this case we might want to.....
*/
void init_spy_join()
{
	struct Channel *c;

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
	time_t now = time(NULL);

        sts(":%s SJOIN %ld %s + :@%s", settings.name,
                now, SPYCHAN, AdminServ.nick);
	c = findchannel(SPYCHAN);
        if (c == NULL)
                c = AddChannel(SPYCHAN, now);
#elif defined(IRCD_2_10)
        /* Here is the only one that we hack ops... */
        sts(":%s NJOIN %s :@%s", settings.name, SPYCHAN, AdminServ.nick);
        c = findchannel(SPYCHAN);
        if (c == NULL)
                c = AddChannel(SPYCHAN);
#endif
	AddUserFlagToChannel(c, AdminServ.u, CMODE_OP);
}
#endif

void InitMessageStrings()
{
#ifdef SUPPORT_MSG_UH
        /* Create the user@host that we are to save some cycles later...
        ** IF this features is used often
        */
#ifdef STATSERV
        snprintf(s_UH, (MAXHOST + MAXUSER + 2), "%s@%s", StatServ.user,
                settings.name);
#endif
#ifdef SPLITSERV
        snprintf(sp_UH, (MAXHOST + MAXUSER + 2), "%s@%s", SplitServ.user,
                settings.name);
#endif
        snprintf(a_UH, (MAXHOST + MAXUSER + 2), "%s@%s", AdminServ.user,
                settings.name);
#ifdef JUPESERV
        snprintf(j_UH, (MAXHOST + MAXUSER + 2), "%s@%s", JupeServ.user,
                settings.name);
#endif
#endif

#ifdef SUPPORT_MSG_OPERS
        /* Same as above...If we call it once, we'll definately call it
        ** again...
        */
        snprintf(s_OPERS, (MAXHOST + 7), "opers@%s", settings.name);
#endif
}
