/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** myusers.c - Functions to track logged in users
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: myusers.c,v 1.7 2002/11/08 15:21:08 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"

dlink_list myuhead;

void LoadMyUsers()
{
	FILE *fp;
	int parc;
	int i;
	char *parv[MAXPARA];
	char buf[BUFSIZE];
	struct MyUser *myu;
	dlink_node *dl;

        /* Clear the parv */
        for (i = 0; i < MAXPARA; i++)
                parv[i] = NULL;

	if ((fp = fopen(USERFILE, "r")) == NULL)
	{
		fprintf(stderr, "Please type this: echo \"<nickname> "
			"<password> 0 >> %s\"\n", USERFILE);
		fprintf(stderr, "Replace nickname with your nickname, and "
			"password with your password, encrypting if "
			"necessary\n");
		fprintf(stderr, "This is used for the LOGIN feature of "
			"Stats.\n");
		log("Unable to open userlist.");
		exit(0);
	}

	while (fgets(buf, BUFSIZE, fp))
	{
		parc = msgtoken(buf, parv);

		if (parc < 2)
		{
			fprintf(stderr, "Invalid entry in user file: %s\n",
				USERFILE);
			fprintf(stderr, "Format: nick pass 0\n");
			fprintf(stderr, "Or: nick pass <time last seen>\n");
			fprintf(stderr, "Exiting...\n");
			exit(0);
		}

                myu = smalloc(sizeof(struct MyUser));
		memcounts.myuser++;
                strlcpy(myu->username, parv[0], MAXNICK + 1);
                strlcpy(myu->password, parv[1], LINE);

		if (parc < 3)
		{
			log("Invalid time value in user file %s, defaulting "
				"to 0", USERFILE);
			/* smalloc'ed ram */
		}
		else
		{
			myu->lastseen = atol(parv[2]);
		}

		dl = dlink_create();
		dlink_add(myu, dl, &myuhead);
	}
	fclose(fp);
}

void SaveMyUsers()
{
	struct MyUser *myu;
	dlink_node *dl;
	FILE *fp;

	if (myuhead.head == NULL)
		return;

	if ((fp = fopen(USERFILE, "w")) == NULL)
	{
		log("Unable to open %s for writing.", USERFILE);
		return;
	}
	DLINK_FOREACH(dl, myuhead.head)
	{
		myu = (struct MyUser *) dl->data;
		fprintf(fp, "%s %s %ld\n", myu->username, myu->password,
			myu->lastseen);
	}
	fclose(fp);
}

void DeleteMyUser(char *username)
{
	struct MyUser *myu;
	dlink_node *dl;

	if (myuhead.head == NULL)
		return;

	DLINK_FOREACH(dl, myuhead.head)
	{
		myu = (struct MyUser *) dl->data;
		if (!strcasecmp(myu->username, username))
		{
			dlink_delete(dl, &myuhead);
			dlink_free(dl);
			free(myu);
			memcounts.myuser--;
			return;
		}
	}
}

struct MyUser *findmyuser(char *nick)
{
	struct MyUser *myu;
	dlink_node *dl;

	DLINK_FOREACH(dl, myuhead.head)
	{
		myu = (struct MyUser *) dl->data;
		if (!strcasecmp(nick, myu->username))
			return myu;
	}
	return NULL;
}
