/* Stats - IRC Statistical Services for Hybrid, csircd, and IRCnet 2.10
** statserv.c - Statistical service interface
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: statserv.c,v 1.132 2002/11/23 23:05:03 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "match.h"

#ifdef STATSERV

static char errstr[BUFSIZE];

static int s_server(struct User *, int, char **);
static int s_map(struct User *, int, char **);
static int s_version_map(struct User *, int, char **);
static int s_uptime(struct User *, int, char **);
static int s_netstats(struct User *, int, char **);
static int s_daily(struct User *, int, char **);
static int s_stats(struct User *, int, char **);
static int s_country(struct User *, int, char **);
static int s_tld(struct User *, int, char **);
static int s_tld_map(struct User *, int, char **);
static int s_operlist(struct User *, int, char **);
static int s_join(struct User *, int, char **);
static int s_cping(struct User *, int, char **);
static int s_cversion(struct User *, int, char **);
static int s_update(struct User *, int, char **);
static int s_url(struct User *, int, char **);
static int s_status(struct User *, int, char **);
static int s_help(struct User *, int, char **);

static struct MsgTable stat_table[] =
{
	{ "HELP",	0,0,NULL,	s_help },
	{ "SERVER",	0,0,NULL,	s_server },
	{ "MAP",	0,0,NULL,	s_map },
	{ "VERSIONMAP",	0,0,NULL,	s_version_map },
	{ "UPTIME",	0,0,NULL,	s_uptime },
	{ "NETSTATS",	0,0,NULL,	s_netstats },
	{ "DAILY",	0,0,NULL,	s_daily },
	{ "COUNTRY",    0,0,NULL,       s_country },
	{ "TLD",	0,0,NULL,	s_tld },
	{ "TLDMAP",	0,0,NULL,	s_tld_map },
	{ "OPERLIST",	0,0,NULL,	s_operlist },
	{ "STATS",	0,0,NULL,	s_stats },
	{ "JOIN",	0,0,NULL,	s_join },
	{ "\001PING",	0,0,NULL,	s_cping },
	{ "\001VERSION\001",	0,0,NULL,	s_cversion },
	{ "UPDATE",	0,0,NULL,	s_update },
	{ "URL",	0,0,NULL,	s_url },
	{ "STATUS",	0,0,NULL,	s_status },
	{ (char *)0,	0,0,NULL,	(int (*)()) 0 }
};

static struct HelpTable help_table[] =
{
        { "SERVER",     "help/statserv/server",         0 },
        { "MAP",        "help/statserv/map",            0 },
        { "NETSTATS",   "help/statserv/netstats",       0 },
        { "DAILY",      "help/statserv/daily",          0 },
	{ "COUNTRY",    "help/statserv/country",        0 },
        { "TLD",        "help/statserv/tld",            0 },
        { "TLDMAP",     "help/statserv/tldmap",         0 },
        { "OPERLIST",   "help/statserv/operlist",       0 },
        { "STATS",      "help/statserv/stats",          1 },
        { "JOIN",       "help/statserv/join",           1 },
        { "UPDATE",     "help/statserv/update",         1 },
        { "URL",        "help/statserv/url",            0 },
        { "STATUS",     "help/statserv/status",         0 },
        { "UPTIME",     "help/statserv/uptime",         0 },
        { "VERSIONMAP", "help/statserv/versionmap",     0 },
        { (char *)0,    (char *)0,                      0 }
};

#ifdef IRCD_2_10
void statserv(struct User *u, char *target, char *param, int is_squery)
#else
void statserv(struct User *u, char *target, char *param)
#endif
{
	char *parv[MAXPARA];
	int parc = 0;
	int i;
	int status;
	static char buf[BUFSIZE];

        /* Clear the parv */
        for (i = 0; i < MAXPARA; i++)
                parv[i] = NULL;

	/* Clear the error string */
	errstr[0] = '\0';

	strlcpy(buf, param, BUFSIZE);

	if (param != NULL)
		parc = msgtoken(param, parv);

	netinfo.requests++;

        if (!is_channel(target) && flood(u))
                return;

	if (parc == 0)
		return;

	if (parv[0][0] != '\001')
		log("%s received message from %s: %s", StatServ.nick, 
			u->nick, buf);

        if (IsOnlyOpers() && !IsOper(u))
	{
		if (parv[0][0] != '\001')
		{
                	privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
                       		 "This service is only available to IRCops.");
                	return;
		}
        }

	/* All the preliminary junk is done, now we have to go through 
	** the tables and call the handler for that command
	*/
        for (i=0; stat_table[i].handler; i++)
        {
                if (!strcasecmp(parv[0], stat_table[i].name))
                {
                    status = stat_table[i].handler(u, parc, parv);
                    if (status == 1)
                        log("Error with statserv command %s: %s", parv[0],
			(errstr) ? errstr : "No error message");
		    /* Matched a value, done, then return */
		    return;
                }
        }
	/* If we get here, we did not get a valid command...
	** Make sure not to say anything to the user if it was a 
	** CTCP that was not matched
	*/
	if (*parv[0] == '\001')
		return;
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Invalid command: %s",
		parv[0]);
	return;
}

/* In all of these functions, parv[0] contains the command, followed
** by the parameters
*/

static int s_help(struct User *u, int parc, char *parv[])
{
	int i;

        spy("%s:  HELP Request by %s for %s", StatServ.nick, u->nick,
                (parv[1]) ? parv[1] : "NO QUERY");
	sendto_dcc_flag(CF_COMMANDS, "%s:  HELP Request by %s for %s",
		StatServ.nick, u->nick, (parv[1]) ? parv[1] : "NO QUERY");

	if (parc < 2)
	{
		send_help_file(u->nick, StatServ.nick, MSG_DEFAULT,
			"help/statserv/index", "index");
		if (u->myuser != NULL)
			send_help_file(u->nick, StatServ.nick, MSG_DEFAULT,
				"help/statserv/aindex", "aindex");
		return 0;
	}

	for (i = 0; help_table[i].filename; i++)
	{
		if (!strcasecmp(parv[1], help_table[i].name))
                {
			if (help_table[i].myuonly)
			{
				if (u->myuser)
					send_help_file(u->nick, StatServ.nick,
						MSG_DEFAULT,
						help_table[i].filename,
						help_table[i].name);
				else
					privmsg(u->nick, StatServ.nick,
						MSG_DEFAULT,
						"Permission Denied.");
			}
			else
			{
				send_help_file(u->nick, StatServ.nick,
					MSG_DEFAULT, help_table[i].filename,
					help_table[i].name);
			}

			return 0;
		}
	}

	privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"Unknown Help Topic: %s", parv[1]);
	snprintf(errstr, BUFSIZE, "Unknown HELP topic: %s", parv[1]);
	return 1;
}

static int s_server(struct User *u, int parc, char *parv[])
{
	struct Server *s = NULL;
	struct Stats *st = NULL;

	if (parc < 2)
	{
		privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Syntax: /msg %s SERVER <server name>", StatServ.nick);
		strcpy(errstr, "Missing parameter to SERVER");
		return 1;
	}

	s = findserver(parv[1]);
	st = findstats(parv[1]);

	spy("%s: SERVER %s Request by %s", StatServ.nick, parv[1], u->nick);
	sendto_dcc_flag(CF_COMMANDS, "%s: SERVER %s Request by %s",
		StatServ.nick, parv[1], u->nick);

	if (!st)
	{
		privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"No information available on %s", parv[1]);
		return 0;
	}

        /* If we want the server to be hidden, catch it here */
        if (s && IsSHide(s) && !IsOper(u))
        {
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
                    "No information available on %s (hidden server)", parv[1]);
                return 0;
        }

        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"Statistics for %s since %s", parv[1], sftime(st->starttime));
        if (!s)
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Server Last Seen:  %s", sftime(st->lastseen));
        if (s)
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Current Users: %-3ld (%.0f%%)", s->users,
                        (float)s->users / (float)netinfo.users * 100);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Maximum Users: %-3ld %s",
		st->maxusers, sftime(st->t_maxusers));
        if (s)
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Current Opers: %d",
			s->opers);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Maximum Opers: %-3d %s",
		st->maxopers,
                sftime(st->t_maxopers));
#ifdef IRCD_2_10
	if (s)
		privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Current Services: %d", s->services);
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Maximum Services: %-3d %s",
		st->maxservices, sftime(st->t_maxservices));
#endif
#if defined(IRCD_HYB7) || defined(IRCD_2_10) || defined(IRCD_BAHAMUT) || \
    defined(IRCD_HYBSVS)
        if (s)
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
                        "Current Away: %d", s->away);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Maximum Away: %-3d %s",
                st->maxaway, sftime(st->t_maxaway));
#endif
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "  IRCop Kills: %d",
		st->operkills);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, " Server Kills: %d",
		st->serverkills);

#if defined(IRCD_HYB7) || defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
	if ((s != NULL) && IsSetEOB(s))
	{
		privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			" EOB Received, burst time: %d seconds",
			s->eob_time - s->connected_since);
	}
#endif

        if (st->numsplits > 0)
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"%s has split from the network %d time%s.",
                st->name, st->numsplits, (st->numsplits == 1) ? "" : "s");
        else
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"%s has never split from the network.", st->name);
	if (s && IsSHide(s))
		privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Server is marked as hidden.");
#ifdef TRACK_VERSION
	if (s)
	{
	        if (s->version[0])
        	        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"Version: %s", s->version);
	        else
        	        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"Version Unknown");
	}
#endif
#ifdef TRACK_ADMIN
	if (s)
	{
		if (s->admin1[0] != '\0')
			privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"Admin Info: %s", s->admin1);
		if (s->admin2[0] != '\0')
                        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"Admin Info: %s", s->admin2);
		if (s->admin3[0] != '\0')
                        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"Admin Info: %s", s->admin3);
	}
#endif
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "End of Statistics.");

	return 0;
}

static int s_version_map(struct User *u, int parc, char *parv[])
{
#ifdef TRACK_VERSION
        struct Server *s = NULL;
        int i;

	spy("%s: VERSIONMAP Request by %s", StatServ.nick, u->nick);
	sendto_dcc_flag(CF_COMMANDS, "%s: VERSIONMAP Request by %s",
		StatServ.nick, u->nick);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
                "[Server]                      [Version]");
        for (i = 0; i < S_TABLE_SIZE; i++)
	{
                for (s = serverlist[i]; s; s = s->next)
		{
                    if (!IsSHide(s) || IsOper(u))
		    {
                            privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
                                "%-30s %s", s->name,
				(s->version && s->version[0]) ?  s->version :
				"Unknown Version");
                    }
                }
        }
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "End of Version Map.");

	return 0;
#else
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "VERSIONMAP Disabled.");
	return 0;
#endif
}

static int s_map(struct User *u, int parc, char *parv[])
{
        struct Server *s;
        int i;

	spy("%s: MAP Request by %s", StatServ.nick, u->nick);
	sendto_dcc_flag(CF_COMMANDS, "%s: MAP Request by %s", StatServ.nick,
		u->nick);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"Network Map [NAME] [USERS/MAX] "
                "[OPERS/MAX] [PERCENT]");
        for (i = 0; i < S_TABLE_SIZE; i++)
	{
                for (s = serverlist[i]; s; s = s->next)
		{
                    if (!IsSHide(s) || IsOper(u))
		    {
#ifdef LARGE_NET
                            privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
                                "%-30s [%3d/%3d] [%2d/%2d] [%2.0f%%]",
#else
                            privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
                                "%-30s [%2d/%2d] [%2d/%2d] [%2.0f%%]",
#endif /* LARGE_NET */
                                s->name, s->users, s->stats->maxusers,
                                s->opers, s->stats->maxopers,
                                (float) s->users / (float) netinfo.users * 100
                                );
                        }
                }
        }
#ifdef LARGE_NET
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
                "%-30s [%4d/%4d] [%3d/%3d] [100%%]",
#else /* LARGE_NET */
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
                "%-30s [%2d/%2d] [%2d/%2d] [100%%]",
#endif /* LARGE_NET */
                 "Totals:", netinfo.users, netinfo.maxusers, netinfo.opers, netinfo.maxopers);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "End of Map.");

	return 0;
}

static int s_status(struct User *u, int parc, char *parv[])
{

	spy("%s: STATUS Request by %s", StatServ.nick, u->nick);
	sendto_dcc_flag(CF_COMMANDS, "%s: STATUS Request by %s",
		StatServ.nick, u->nick);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Statistics Status:");
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Reconnect Time: %d",
		settings.r_time);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Oper Only: %s",
                IsOnlyOpers() ? "Enabled" : "Disabled");
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "HTML Dump: %s",
                IsHtml() ? "Enabled" : "Disabled");
        if (IsHtml())
	{
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "HTML File: %s",
			settings.html);
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"HTML File created every %d seconds",
			settings.html_time);
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "URL: %s",
			settings.url);
        }
        /* settings.name, settings.desc, settings.uplink all redundant, as
	** they can be seen from LINKS or from whois StatServ.nick
        */
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Set to oper: %s",
                IsSetOper() ? "Enabled" : "Disabled");

    if (u->myuser) {
#ifdef ENABLE_QUOTE
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "QUOTE: Enabled");
#endif
        if (settings.flood)
	{
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
                        "Flood Control: %d requests in %d seconds",
			settings.flood, settings.flood_time);
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Flood Action: %s",
                        IsFlood() ? "IGNORE" : "KILL");
        }
    }

    return 0;
}

static int s_uptime(struct User *u, int parc, char *parv[])
{
	spy("%s: UPTIME Request by %s", StatServ.nick, u->nick);
	sendto_dcc_flag(CF_COMMANDS, "%s: UPTIME Request by %s",
		StatServ.nick, u->nick);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Statistics up %s",
		timediff(time(NULL) - me.t_start));
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Statistic Requests: %d",
		netinfo.requests);
	return 0;
}

static int s_netstats(struct User *u, int parc, char *parv[])
{
	spy("%s: NETSTATS Request by %s", StatServ.nick, u->nick);
	sendto_dcc_flag(CF_COMMANDS, "%s: NETSTATS Request by %s",
		StatServ.nick, u->nick);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Network Statistics:");
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"   Current Users: %ld", netinfo.users);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"   Maximum Users: %-2ld %s", netinfo.maxusers,
		sftime(netinfo.t_maxusers));
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "   Current Opers: %d",
		netinfo.opers);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "   Maximum Opers: %-2d %s",
		netinfo.maxopers, sftime(netinfo.t_maxopers));
#if defined(IRCD_HYB7) || defined(IRCD_2_10) || defined(IRCD_BAHAMUT) || \
    defined(IRCD_HYBSVS)
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "   Users Set Away: %-2d",
		netinfo.away);
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"   Maximum Away: %-2d %s", netinfo.maxaway,
		sftime(netinfo.t_maxaway));
#endif
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "   Current Servers: %d",
		netinfo.servers);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"   Maximum Servers: %-2d %s", netinfo.maxservers,
		sftime(netinfo.t_maxservers));
#ifdef IRCD_2_10
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "   Current Services: %d",
		netinfo.services);
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"   Maximum Services: %-2d %s",
		netinfo.maxservices, sftime(netinfo.t_maxservices));
#endif
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "   Current Channels: %d",
		netinfo.channels);
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"   Maximum Channels: %-2d %s",
		netinfo.maxchannels, sftime(netinfo.t_maxchannels));
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "End of Information.");

	return 0;
}

static int s_daily(struct User *u, int parc, char *parv[])
{
	spy("%s: DAILY Request by %s", StatServ.nick, u->nick);
	sendto_dcc_flag(CF_COMMANDS, "%s: DAILY Request by %s",
		StatServ.nick, u->nick);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"Daily Network Statistics:");
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, " ");
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"  Maximum Servers: %-2d %s", daily.servers,
		sftime(daily.t_servers));
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "  Maximum Users: %-2d %s",
		daily.users, sftime(daily.t_users));
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"  Maximum Opers: %-2d %s", daily.opers,
                sftime(daily.t_opers));
#ifdef IRCD_2_10
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"  Maximum Services: %-2d %s", daily.services,
		sftime(daily.t_services));
#endif
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"  Maximum Channels: %-2d %s", daily.channels,
		sftime(daily.t_channels));
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"All daily statistics are reset at midnight.");
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "End of Information.");

	return 0;
}

static int s_country(struct User *u, int parc, char *parv[])
{
	int i;
	struct TLD *t;
	char search[LINE];
	int matches = 0;

	if (parv[1] == NULL)
	{
		privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Syntax: /msg %s COUNTRY <countryname>", StatServ.nick);
		privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"For additional help, /msg %s HELP", StatServ.nick);
		strcpy(errstr, "Missing parameter in COUNTRY");
		return 1;
	}

	/* Wrap the entered search string with *'s */
	snprintf(search, LINE, "*%s*", parv[1]);

        for (i = 0; i < T_TABLE_SIZE; i++)
        {
                for(t = tldlist[i]; t; t = t->next)
                {
			if (matches == COUNTRY_LIMIT)
			{
				privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
					"Too many matches, only %d returned",
					matches);
				return 0;
			}
                        if (match(search, t->country) == 0)
			{
				matches++;
		                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
					"Match %d:  %s (%s)", matches,
					t->country, t->tld);
                		if (t->native != NULL)
                        		privmsg(u->nick, StatServ.nick,
						MSG_DEFAULT, "Native Name: %s",
						t->native);
				/* There can be multiple matches for the
				** same strings, consider Korea or Congo
				*/
			}
                }
        }

	return 0;
}

static int s_tld(struct User *u, int parc, char *parv[])
{
        struct TLD *tmp;
	char *tld = parv[1];

        if (!tld)
	{
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Syntax: /msg %s TLD <tld>", StatServ.nick);
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"For additional help, /msg %s HELP", StatServ.nick);
		strcpy(errstr, "Missing parameter in TLD");
                return 1;
        }

        if (*tld == '*')
                tld++;
        if (*tld == '.')
                tld++;

	/* Catch it when it's just * or . */
	if (*tld == '\0')
	{
		privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Syntax: /msg %s TLD <tld>", StatServ.nick);
		privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"For additional help, /msg %s HELP", StatServ.nick);
		strcpy(errstr, "Missing parameter in TLD");
		return 1;
	}

	spy("%s: TLD %s Request by %s", StatServ.nick, tld, u->nick);
	sendto_dcc_flag(CF_COMMANDS, "%s: TLD %s Request by %s", StatServ.nick,
		tld, u->nick);

        tmp = findtld(tld);

        if (tmp == NULL)
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Top Level Domain %s does not exist.", tld);
        else
	{
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "%s (%s)",
			tmp->country, tmp->tld);
		if (tmp->native != NULL)
			privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"Native Name: %s", tmp->native);
	}

	return 0;
}

static int s_tld_map(struct User *u, int parc, char *parv[])
{
        struct TLD *t;
	int i;

	spy("%s: TLDMAP Request by %s", StatServ.nick, u->nick);
	sendto_dcc_flag(CF_COMMANDS, "%s: TLDMAP Request by %s",
		StatServ.nick, u->nick);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"Top Level Domain Statistics:");
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"   TLD Number / Percent  Country");
	privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
		"   --- ------   -------  -------");

        for (i = 0; i < T_TABLE_SIZE; i++)
        {
                for(t = tldlist[i]; t; t = t->next)
                {
			if (t->users != 0)
				privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"%9s %3d (%2.0f%%) -> %s", t->tld,
				t->users,
				(float)t->users / (float)netinfo.users * 100,
				t->country);
                }
        }
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "End of Information.");

	return 0;
}

static int s_operlist(struct User *u, int parc, char *parv[])
{
	dlink_node *dl;
        int j = 0;
        int away = 0;
        struct User *ul;
        char *server = parv[2];

        spy("%s: OPERLIST Request by %s", StatServ.nick, u->nick);
	sendto_dcc_flag(CF_COMMANDS, "%s: OPERLIST Request by %s",
		StatServ.nick, u->nick);
        if (parv[1] && !strcasecmp(parv[1], "NOAWAY"))
        {
                away = 1;
        }
        if (!away && parv[1] && strchr(parv[1], '.'))
                server = parv[1];

        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "On-Line IRCops:");

	DLINK_FOREACH(dl, global_oper_list.head)
        {
		ul = (struct User *) dl->data;
                if (away && IsAway(ul))
                        continue;
                if (IsSHide(ul->server) && !IsOper(u))
                        continue;
                if (!IsOper(ul))
                        continue;
                if (!server)
                {
                        j++;
                        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
#ifdef OPER_MODES
                                "[%2d] [%c] %-15s %s", j, opermode(ul),
                                ul->nick,
#else
                                "[%2d] %-15s %s",j, ul->nick,
#endif
                                ul->server->name);
                        continue;
                }
                else
                {
                        if (strcasecmp(server, u->server->name))
                                continue;
                        j++;
                        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
#ifdef OPER_MODES
                                "[%2d] [%c] %-15s %s", j, opermode(ul),
#else
                                "[%2d] %-15s %s", j,
#endif
                                ul->nick, ul->server->name);
                        continue;
                }
        }
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "End of Listing.");

        return 0;
}

static int s_stats(struct User *u, int parc, char *parv[])
{
        struct Stats *st;

        if (!u->myuser)
	{
                log("Access Denied (STATS) to %s", u->nick);
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Access Denied.");
		/* Not a code error */
                return 0;
        }

        if (parc < 2)
	{
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Syntax: /msg %s STATS [DEL|LIST|COPY]",
                        StatServ.nick);
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"For additional help, /msg %s HELP", StatServ.nick);
		strcpy(errstr, "Missing parameter in STATS");
                return 1;
        }
        if (!strcasecmp(parv[1], "LIST"))
	{
                int i = 1;
		dlink_node *dl;

		spy("%s: STATS LIST Request by %s", StatServ.nick, u->nick);
		sendto_dcc_flag(CF_COMMANDS, "%s: STATS LIST Request by %s",
			StatServ.nick, u->nick);
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Statistics Database:");
		DLINK_FOREACH(dl, shead.head)
		{
			st = (struct Stats *) dl->data;
                        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"[%-2d] %s", i, st->name);
                        i++;
                }
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "End of List.");
                log("%s requested STATS LIST.", u->nick);
        }
	else if (!strcasecmp(parv[1], "DEL"))
	{
		dlink_node *dl;

                if (parc < 3)
		{
                        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"Syntax: /msg %s STATS DEL <name>",
				StatServ.nick);
                        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"For additonal help, /msg %s HELP",
				StatServ.nick);
			strcpy(errstr, "Missing parameter in STATS");
                        return 1;
                }
                st = findstats(parv[2]);
                if (!st)
		{
			spy("%s: STATS DEL %s Attempt by %s (failed)",
				StatServ.nick, parv[2], u->nick);
			sendto_dcc_flag(CF_COMMANDS,
				"%s: STATS DEL %s Attempt by %s (failed)",
				StatServ.nick, parv[2], u->nick);
                        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"%s is not in the database!", parv[2]);
			snprintf(errstr, BUFSIZE, "Unknown server: %s",
				parv[2]);
                        return 1;
                }
		DLINK_FOREACH(dl, shead.head)
		{
			st = (struct Stats *) dl->data;
			if (!strcasecmp(parv[2], st->name))
			{
				dlink_delete(dl, &shead);
				dlink_free(dl);
				free(st);
				memcounts.stats--;
				break;
			}
		}
		spy("%s: STATS DEL %s Request by %s", StatServ.nick, parv[2],
			u->nick);
		sendto_dcc_flag(CF_COMMANDS, "%s: STATS DEL %s Request by %s",
			StatServ.nick, parv[2],
			u->nick);
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Removed %s from the database.", parv[2]);
                log("%s requested STATS DEL %s", u->nick, parv[2]);
        }
	else if (!strcasecmp(parv[1], "COPY"))
	{
                struct Server *s;
		struct Stats *st2;
		dlink_node *dl;

                if (parc < 4)
		{
                        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"Syntax: /msg %s STATS COPY <name> <newname>",
				StatServ.nick);
			strcpy(errstr, "Missing parameter in STATS");
                        return 1;
                }

                st = findstats(parv[2]);
                if (!st)
		{
                        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"No entry in the database for %s", parv[2]);
			snprintf(errstr, BUFSIZE, "Unknown stats entry: %s",
				parv[2]);
                        return 1;
                }
                s = findserver(parv[2]);
                if (s)
		{
                        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
				"Server %s is online!", parv[2]);
			snprintf(errstr, BUFSIZE, "Server %s is online",
				parv[2]);
                        return 1;
                }

		/* Wait until we are sure before removing a statistics
		** entry
		*/
		DLINK_FOREACH(dl, shead.head)
		{
			st2 = (struct Stats *) dl->data;
			if (!strcasecmp(parv[3], st2->name))
			{
				dlink_delete(dl, &shead);
				dlink_free(dl);
				free(st2);
				memcounts.stats--;
				break;
			}
		}

                s = NULL;
                strlcpy(st->name, parv[3], sizeof(st->name));
                s = findserver(st->name);
                if (s)
                        s->stats = st;
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Moved database entry for %s to %s", parv[2],
			parv[3]);
                log("%s requested STATS COPY %s -> %s", u->nick, parv[2],
			parv[3]);
		spy("%s: STATS COPY %s %s Request by %s", StatServ.nick,
			parv[2], parv[3], u->nick);
		sendto_dcc_flag(CF_COMMANDS,
			"%s: STATS COPY %s %s Request by %s", StatServ.nick,
			parv[2], parv[3], u->nick);
        }
	else
	{
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Invalid Argument.");
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"For help, /msg %s HELP", StatServ.nick);
		strcpy(errstr, "Invalid argument in STATS");
		return 1;
        }

	SaveStats();
	return 0;
}

static int s_join(struct User *u, int parc, char *parv[])
{
        if (!u->myuser)
	{
                log("Access Denied (JOIN) to %s", u->nick);
		spy("%s: JOIN Attempt by %s (failed)", StatServ.nick, u->nick);
		sendto_dcc_flag(CF_COMMANDS,
			"%s: JOIN Attempt by %s (failed)", StatServ.nick,
			u->nick);
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Access Denied.");
		/* Not a code error */
                return 0;
        }

        if (parv[1])
	{
                if ( !is_channel(parv[1]) )
		{
                        privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
                                "Please enter a valid channel");
			snprintf(errstr, BUFSIZE, "Invalid channel: %s",
				parv[1]);
                        return 1;
                }

#if defined(IRCD_HYBRID) || defined(IRCD_HYB6) || defined(IRCD_HYB7) || \
    defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)
                        sts(":%s SJOIN %ld %s + :@%s", settings.name,
				time(NULL), parv[1], StatServ.nick);
#else
#if defined(IRCD_2_10)
                        sts(":%s NJOIN %s :@%s", settings.name, parv[1],
				StatServ.nick);
#endif
#endif

                log("%s!%s@%s made me join %s", u->nick, u->username,
                        u->hostname, parv[1]);
		spy("%s: JOIN %s Request by %s", StatServ.nick, parv[1],
			u->nick);
		sendto_dcc_flag(CF_COMMANDS, "%s: JOIN %s Request by %s",
			StatServ.nick, parv[1], u->nick);

                return 0;
        }
	else
	{
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"Please enter a valid channel");
		strcpy(errstr, "NULL channel attempt in JOIN");
                return 1;
        }
}

static int s_cping(struct User *u, int parc, char *parv[])
{
        char command[BUFSIZE];

	ExpandParv(command, BUFSIZE, 1, parc, parv);
        if (command[0] == '\0')
                return 1;
        sts(":%s NOTICE %s :\001PING %s", StatServ.nick, u->nick, command);
        log("CTCP PING from %s!%s@%s", u->nick, u->username, u->hostname);

        return 0;
}

static int s_cversion(struct User *u, int parc, char *parv[])
{
        sts(":%s NOTICE %s :\001VERSION %s %s %s -> %s\001", StatServ.nick,
             u->nick, stats_version, settings.name, VersionFlags(),
	     stats_date);
        log("CTCP VERSION from %s!%s@%s", u->nick, u->username, u->hostname);

	return 0;
}

static int s_update(struct User *u, int parc, char *parv[])
{
        if (!u->myuser)
	{
                log("Access Denied (UPDATE) to %s", u->nick);
		spy("%s: UPDATE Attempt by %s (failed)", StatServ.nick, u->nick);
		sendto_dcc_flag(CF_COMMANDS,
			"%s: UPDATE Attempt by %s (failed)", StatServ.nick,
			u->nick);
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "Access Denied.");
		/* Not a code error */
                return 0;
        }

        if (!IsHtml())
	{
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"HTML Update Disabled");
                return 0;
        }

        DumpHTML();
	spy("%s: UPDATE Request by %s", StatServ.nick, u->nick);
	sendto_dcc_flag(CF_COMMANDS, "%s: UPDATE Request by %s", StatServ.nick,
		u->nick);
        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "HTML Update Forced");

        log("%s!%s@%s forces an HTML Update", u->nick, u->username,
		u->hostname);
	return 0;
}

static int s_url(struct User *u, int parc, char *parv[])
{
        if (!IsHtml())
	{
                privmsg(u->nick, StatServ.nick, MSG_DEFAULT,
			"HTML Dump Disabled");
                return 0;
        }

        privmsg(u->nick, StatServ.nick, MSG_DEFAULT, "HTML Status URL: %s",
		settings.url);
	spy("%s: URL Request by %s", StatServ.nick, u->nick);
	sendto_dcc_flag(CF_COMMANDS, "%s: URL Request by %s", StatServ.nick,
		u->nick);
        log("%s!%s@%s Looks up URL", u->nick, u->username, u->hostname);

	return 0;
}

#endif
